package cz.mendelu.dnaAnalyser.analyse.cpg;

import cz.mendelu.dnaAnalyser.analyse.zdna.ZdnaAnalyseResult;
import cz.mendelu.dnaAnalyser.jwt.JwtAuthenticationUtils;
import cz.mendelu.dnaAnalyser.jwt.JwtTokenService;
import cz.mendelu.dnaAnalyser.sequence.Sequence;
import cz.mendelu.dnaAnalyser.sequence.SequenceService;
import cz.mendelu.dnaAnalyser.exporter.ExporterService;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceData;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceDataRepository;
import cz.mendelu.dnaAnalyser.utils.analyse.service.heatmap.Heatmap;
import cz.mendelu.dnaAnalyser.utils.controller.DeleteMapping;
import cz.mendelu.dnaAnalyser.utils.controller.GetMapping;
import cz.mendelu.dnaAnalyser.utils.controller.*;
import cz.mendelu.dnaAnalyser.utils.controller.PostMapping;
import cz.mendelu.dnaAnalyser.utils.controller.PutMapping;
import cz.mendelu.dnaAnalyser.utils.swagger.ApiPaginationParams;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import lombok.Builder;
import lombok.Data;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.http.ContentDisposition;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.core.Authentication;
import org.springframework.web.bind.annotation.*;
import springfox.documentation.annotations.ApiIgnore;
import cz.mendelu.dnaAnalyser.utils.analyse.data.ExporterEnums.Analysis;
import cz.mendelu.dnaAnalyser.utils.analyse.data.ExporterEnums.Output;

import javax.validation.Valid;
import javax.validation.constraints.NotNull;

import java.util.ArrayList;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;

import lombok.extern.slf4j.Slf4j;

@Slf4j
@RestController
@Api(tags = "api/analyse/cpg")
@RequestMapping("api/analyse/cpg")
public class CpgController {

    @Data
    static class CpgRequest {
        @NotNull
        private UUID sequence;
        private Set<String> tags;

        private Integer minWindowSize;
        private Double minGcPercentage;
        private Double minObservedToExpectedCpG;
        private Integer minIslandMergeGap;
        private Character firstNucleotide;
        private Character secondNucleotide;
    }

    @Data
    static class HeatMapRequest {
        private Integer sequenceLength;
        private UUID sequenceId;
    }

    @Autowired
    private ExporterService exporterService;

    @Autowired
    private SequenceService sequenceService;

    @Autowired
    private JwtTokenService jwtTokenService;

    @Autowired
    private SequenceDataRepository sequenceDataRepository;

    @Autowired
    private CpgService cpgService;

    @PostMapping()
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Analyse sequence for CpG Islands")
    public Response<Cpg> analyzeCpg(@ApiIgnore Authentication authentication, @RequestBody @Valid CpgController.CpgRequest request) {
        Cpg cpg = cpgService.executeCpgAnalysis(
                Cpg.builder()
                        .sequence(sequenceService.findOne(request.getSequence()))
                        .owner(JwtAuthenticationUtils.toUser(authentication))
                        .tags(request.getTags())
                        .minWindowSize(request.getMinWindowSize())
                        .minGcPercentage(request.getMinGcPercentage())
                        .minObservedToExpectedCpG(request.getMinObservedToExpectedCpG())
                        .minIslandMergeGap(request.getMinIslandMergeGap())
                        .firstNucleotide(request.getFirstNucleotide())
                        .secondNucleotide(request.getSecondNucleotide())
                        .build());
        return Response.response(cpg);
    }

    @GetMapping("/{id}/substring")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get a substring of the sequence by ID, start, and end positions.")
    public ResponseEntity<String> getSequenceSubstring(
            @PathVariable UUID id,
            @RequestParam("start") Integer start,
            @RequestParam("end") Integer end) {

        // Fetch the sequence using the provided ID. You might need to adjust this code
        // based on how your SequenceService is implemented and how sequences are stored.
        Sequence sequence = sequenceService.findOne(id);
        if (sequence == null) {
            return new ResponseEntity<>("Sequence not found", HttpStatus.NOT_FOUND);
        }


        String sequence_string = sequenceDataRepository.load(sequence).toPlain();

        // Check if start and end positions are within the sequence length
        if (start < 0 || end > sequence_string.length() || start >= end) {
            return new ResponseEntity<>("Invalid start or end position", HttpStatus.BAD_REQUEST);
        }

        // Extract the substring based on the provided start and end positions
        String substring = sequence_string.substring(start, end);

        // Return the substring
        return new ResponseEntity<>(substring, HttpStatus.OK);
    }


    @GetMapping("/{id}/analysis")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get one CpG analysis by ID.")
    public Response<Cpg> getCpgAnalysis(@PathVariable UUID id) {
        Cpg cpg = cpgService.findOne(id);
        Response<Cpg> response = Response.response(cpg);
        response.setDownloadToken(jwtTokenService.generateDownloadToken());
        return response;
    }

    @GetMapping("/{id}/cpg")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get CpG islands analysis by ID.")
    public ResponsePage<CpgAnalyseResult> getCpgList(@PathVariable UUID id,
                                                     Integer sequenceStart,
                                                     Integer sequenceLength,
                                                     PaginationParam paginationParam,
                                                     SortParam sortParam) {
        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<CpgAnalyseResult> cpgPage = cpgService.returnCpgAnalysis(id, sequenceStart, sequenceLength, pageable);
        return ResponsePage.responsePage(cpgPage);
    }

    @GetMapping("/{id}/average/length")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get average CPG length for single analysis")
    public Integer getAvgCpGLen(@PathVariable UUID id)
    {
        return cpgService.avgCpgLen(id);
    }

    @GetMapping
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiPaginationParams
    @ApiOperation("Get page with CpG analysis.")
    public ResponsePage<Cpg> getPageWithCpgAnalysis(PaginationParam paginationParam, SortParam sortParam, TagFilterParam tagFilterParam) {
        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<Cpg> cpgPage = cpgService.findAll(pageable, tagFilterParam.getTags());
        return ResponsePage.responsePage(cpgPage);
    }

    @GetMapping("/tag")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiPaginationParams
    @ApiOperation("Get all tags defined for sequences")
    public ResponsePage<String> getPageWithTags(PaginationParam paginationParam, SortParam sortParam) {
        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<String> tagPage = cpgService.findAllTags(pageable);
        return ResponsePage.responsePage(tagPage);
    }

    @PutMapping("/{id}/tags")
    @ResponseStatus(HttpStatus.ACCEPTED)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Modify tags")
    public Response<Cpg> modifyTags(@PathVariable UUID id, @RequestBody TagModifyRequest request) {
        Cpg cpg = cpgService.modifyTags(id, request.getTags());
        return Response.response(cpg);
    }

    @DeleteMapping("/{id}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Delete one CpG analysis by ID")
    public void deleteCpgAnalysis(@PathVariable UUID id) {
        cpgService.delete(id);
    }


    @GetMapping(path = "/{id}/cpg.csv", produces = MediaType.TEXT_PLAIN_VALUE)
    @ResponseStatus(HttpStatus.OK)
    @ApiOperation("Get CSV file.")
    public ResponseEntity<String> getCsv(@PathVariable UUID id) {
        String csv = exporterService.exportData(id, Analysis.CPG, Output.CSV);
        HttpHeaders headers = new HttpHeaders();
        headers.setContentDisposition(ContentDisposition.builder("attachment").filename(id + ".csv").build());
        headers.setContentType(MediaType.valueOf("text/csv"));
        headers.setContentLength(csv.length());
        return new ResponseEntity<>(csv, headers, HttpStatus.OK);
    }

    @GetMapping(path = "/{id}/cpg.bedgraph", produces = MediaType.TEXT_PLAIN_VALUE)
    @ResponseStatus(HttpStatus.OK)
    @ApiOperation("Get BEDGRAPH file.")
    public ResponseEntity<String> getBedgraph(@PathVariable UUID id) {
        String bedgraph = exporterService.exportData(id, Analysis.CPG, Output.BEDGRAPH);
        HttpHeaders headers = new HttpHeaders();
        headers.setContentDisposition(ContentDisposition.builder("attachment").filename(id + ".bedgraph").build());
        headers.setContentType(MediaType.valueOf("text/bedgraph"));
        headers.setContentLength(bedgraph.length());
        return new ResponseEntity<>(bedgraph, headers, HttpStatus.OK);
    }

    @GetMapping("/{id}/heatmap")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get CpG heatmap by Id.")
    public Heatmap getCpgHeatmap(
            @PathVariable UUID id,
            Integer segments,
            Integer from,
            Integer to)
    {
        return cpgService.getHeatmap(id, segments, from, to);
    }
}
