package cz.mendelu.dnaAnalyser.analyse.p53predictor;

import cz.mendelu.dnaAnalyser.sequence.nucleotide.Nucleotide;
import cz.mendelu.dnaAnalyser.sequence.stream.Window;
import org.springframework.stereotype.Service;

import java.util.Arrays;
import java.util.List;

import static cz.mendelu.dnaAnalyser.sequence.nucleotide.Nucleotide.*;

/**
 * This Service class provide algorithms for p53predictor analyse.
 */
@Service
public class P53PredictorAnalyse {

    private final double referential = -7.61;
    private final double maxLoss = 7.22; // maximanlni ztrata z tabulky
    private final double affinityThreshold = 2.00; // afinita pod tuto hodnotu se nepočítá ( = 0%)
    private final List<Nucleotide> nucleotideRows = Arrays.asList(A, T, G, C);
    private final double[][] deltaLogKd = {
            /*        G     G     A/G   C     A     T     G     C     C     C     G     G     G     C     A     T     G     T/A   C     C  */
            /*A*/    {0.05, 0.03, 0.00, 0.59, 0.00, 0.16, 0.55, 0.25, 0.15, 0.12, 0.03, 0.10, 0.05, 0.62, 0.00, 0.21, 0.31, 0.15, 0.10, 0.07},
            /*T*/    {0.07, 0.10, 0.15, 0.31, 0.21, 0.00, 0.62, 0.05, 0.10, 0.03, 0.12, 0.15, 0.25, 0.55, 0.16, 0.00, 0.59, 0.00, 0.03, 0.05},
            /*G*/    {0.00, 0.00, 0.00, 0.55, 0.28, 0.35, 0.00, 0.47, 0.29, 0.13, 0.00, 0.00, 0.00, 0.61, 0.32, 0.47, 0.00, 0.40, 0.18, 0.11},
            /*C*/    {0.11, 0.18, 0.40, 0.00, 0.47, 0.32, 0.61, 0.00, 0.00, 0.00, 0.13, 0.29, 0.47, 0.00, 0.35, 0.28, 0.55, 0.00, 0.00, 0.00}
    };

    /**
     * Calculate p53Predictor score for given NucleicBuffer (full length)
     *
     * @param window with sequence with min length 20 nuclide.
     * @return result with Data.
     * @throws IllegalArgumentException if sequence has wrong length.
     */
    public P53PredictorResult calcP53PredictorScore(Window window) {
        if (window.getSize() != 20)
            throw new IllegalArgumentException("Input seqence windows has not length 20.");

        double diff = this.calcDifference(window);

        P53PredictorResult result = new P53PredictorResult();
        result.setPredictor(diff);
        result.setDifference(referential + diff);
        result.setAffinity((diff <= affinityThreshold) ? (1 - (diff / affinityThreshold)) : 0.0);
        result.setSequence(window.toPlain());
        result.setPosition(window.getPosition());
        result.setLength(window.getSize());
        return result;
    }

    private double calcDifference(Window window) {
        return window.doubleStream((pos, nucleotide) -> {
            int row = nucleotideRows.indexOf(nucleotide);
            return this.deltaLogKd[row][pos];
        }).sum();
    }

}
